#include "stdafx.h"





namespace ThinAudioGateway {





/*
################################################################################################

TagDriverC class implementation
================================================================================================

################################################################################################
*/





/*
Constructor
================================================================================================
*/ 

TagDriverC::TagDriverC (GUID const & xClassId)

: ClassId (xClassId)

#if _Debug

  , Dbg (&::Dbg, 0, L"Drv")

#endif // _Debug

{

  *InterfacePath = L'\0';
  Connection = 0;

  Debug (6, L"Constructed, class ID %s", GuidText (ClassId));

  Check ();

}





/*
Destructor
================================================================================================
*/ 

TagDriverC::~TagDriverC () {

  Debug (6, L"Destructing");

  Check ();

  Close ();
  Forget ();

}





/*
Create driver object
================================================================================================
*/ 

HRESULT _stdcall TagDriverC::Create (TagDriver * & Obj, GUID const & xClassId) {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  TagDriverC * Drv = nullptr;

  Obj = nullptr;

  {

#if _Debug

    GblDebug (6, L"Drv: Creating driver object");

#endif _Debug

    Drv = new TagDriverC (xClassId);

    if (Drv == nullptr) {

#if _Debug

      GblDebug (9, L"Drv: No memory for driver object");

#endif _Debug

      Res = HRESULT_FROM_WIN32 (ERROR_NOT_ENOUGH_MEMORY);
      goto Exit;

    }

#if _Debug

    GblDebug (7, L"Drv: Created TagDriver object %#x", Drv);

#endif _Debug

    Drv->Check ();

    Obj = Drv;

    Res = S_OK;

  }

Exit:

  if (FAILED (Res)) {

    if (Drv != nullptr) {

      delete Drv;
      Drv = nullptr;

    }

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





/*
Delete driver object
================================================================================================
*/ 

void TagDriverC::Delete () {

  Debug (6, L"Deleting driver object %#x", this);

  delete this;

}





/*
Find KS filter interface
================================================================================================
*/

HRESULT TagDriverC::Find () {

  Check ();

  HDEVINFO DevList = 0;

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (IsFound ()) {
      
      Debug (9, L"Driver interface already found");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    if (IsOpen ()) {

      Debug (9, L"Driver connection already open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    Debug (6, L"Querying interfaces in class %s", GuidText (ClassId));

    DevList = SetupDiGetClassDevs (
      &ClassId, nullptr, 0,
      DIGCF_DEVICEINTERFACE | DIGCF_PROFILE | DIGCF_PRESENT
    );

    if (DevList == INVALID_HANDLE_VALUE) {

      Res = HRESULT_FROM_SETUPAPI (GetLastError ());

      Debug (9, L"Cannot get class %s interfaces (%s)", GuidText (ClassId), ErrorText (Res));

      goto Exit;

    }

    SP_DEVICE_INTERFACE_DATA IfData;

    ZeroMemory (&IfData, sizeof (IfData));

    IfData.cbSize = sizeof (IfData);

    if (!SetupDiEnumDeviceInterfaces (DevList, nullptr, &ClassId, 0, &IfData)) {

      Res = HRESULT_FROM_SETUPAPI (GetLastError ());

      if (Res == HRESULT_FROM_WIN32 (ERROR_NO_MORE_ITEMS)) {

        Res = HRESULT_FROM_WIN32 (ERROR_NOT_FOUND);

        Debug (9, L"No interfaces found");

      } else {

        Debug (9, L"Cannot get first interface info (%s)", ErrorText (Res));

      }

      goto Exit;

    }

    if (!(IfData.Flags & SPINT_ACTIVE)) {

      Debug (9, L"Interface is disabled");

      Res = HRESULT_FROM_WIN32 (ERROR_RESOURCE_DISABLED);
      goto Exit;

    }

    struct IfDetail : public SP_DEVICE_INTERFACE_DETAIL_DATA {

      wchar_t VarPart [MAX_PATH];
      
    } IfDetData;

    ZeroMemory (&IfDetData, sizeof (IfDetData));
    IfDetData.cbSize = sizeof (SP_DEVICE_INTERFACE_DETAIL_DATA);

    DWORD ReqSize;

    SP_DEVINFO_DATA DevInfo;

    ZeroMemory (&DevInfo, sizeof (DevInfo));
    DevInfo.cbSize = sizeof (DevInfo);

    if (
      !SetupDiGetDeviceInterfaceDetail (
        DevList,
        &IfData,
        &IfDetData,
        sizeof (IfDetData),
        &ReqSize,
        &DevInfo
      )
    ) {

      Res = HRESULT_FROM_SETUPAPI (GetLastError ());

      Debug (9, L"Cannot get interface details (%s)", ErrorText (Res));

      goto Exit;

    }

    LocSize const IfPathLen = static_cast <LocSize> (wcslen (IfDetData.DevicePath));

    if (IfPathLen >= NumArrayElems (InterfacePath)) {

      Debug (9, L"Interface path too long (%u)", IfPathLen);

      Res = HRESULT_FROM_WIN32 (ERROR_BUFFER_OVERFLOW);

      goto Exit;

    }

    wcscpy (InterfacePath, IfDetData.DevicePath);

    Debug (7, L"Interface path: %s", InterfacePath);

    Res = S_OK;

  }

Exit:

  if (DevList != 0) {

    SetupDiDestroyDeviceInfoList (DevList);
    DevList = 0;

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





/*
Open the driver (establish a connection)
================================================================================================
*/ 

HRESULT TagDriverC::Open () {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsFound ()) {
      
      Debug (9, L"Driver interface was not successfully found");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    if (IsOpen ()) {

      Debug (9, L"Driver connection already open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    wchar_t ConnPath [MAX_PATH];

    _snwprintf (ConnPath, NumArrayElems (ConnPath), L"%s\\%s\\driver", GetInterface (), GuidText (GetClassId ()));

    Debug (6, L"Opening driver connection: %s", ConnPath);

    Connection = CreateFile (
      ConnPath, GENERIC_READ | GENERIC_WRITE, 0, nullptr, OPEN_EXISTING, 0, 0
    );

    if (Connection == INVALID_HANDLE_VALUE) {

      Connection = 0;
      Res = HRESULT_FROM_WIN32 (GetLastError ());

      if (Res == HRESULT_FROM_WIN32 (ERROR_FILE_NOT_FOUND)) {

        Debug (9, L"Driver interface %s not found", ConnPath);

      } else {

        Debug (9, L"Cannot open driver interface %s (%s)", ConnPath, ErrorText (Res));

      }

      goto Exit;

    }

    Debug (7, L"Driver connection is open (%#x)", Connection);

    Assert (IsOpen ());

    Res = S_OK;

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





/*
Close the driver
================================================================================================
*/ 

void TagDriverC::Close () {

  Check ();

  if (IsOpen ()) {

    Assert (Connection != 0);

    Debug (6, L"Closing driver connection %#x", Connection);

    CloseHandle (Connection);

    Debug (7, L"Driver connection %#x closed", Connection);

    Connection = 0;

  }

  Check ();

}





/*
Issue a control request
================================================================================================
*/ 

HRESULT TagDriverC::ControlRequest (
  UINT const Func,
  DriverReq & Req, LocSize const ReqSize,
  void * const Data, LocSize const DataSize,
  LocSize * const RetSize
) const {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Driver connection is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    Res = IoControlRequest (Connection, Func, Req, ReqSize, Data, DataSize, RetSize);

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





/*
Perform "get driver information" request
================================================================================================
*/ 

HRESULT TagDriverC::GetInfo (DriverInfo & DI) const {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Driver connection is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    DriverReq Req;

    Debug (6, L"Requesting driver info");

    Res = ControlRequest (IoCtl_Driver_GetInfo, Req, sizeof (Req), &DI, sizeof (DI));

    if (FAILED (Res)) {

      Debug (9, L"Cannot get driver info (%s)", ErrorText (Res));

      goto Exit;

    }

    Debug (
      7,
      L"Driver version: %u.%u.%u.%u, format range: %s, lines: %u",
      DI.VerMajor, DI.VerMinor, DI.VerRev, DI.VerBuild, WfrText (DI.FormatRange), DI.NumLines
    );

    Assert (Res == S_OK);

  }

Exit:

  Check ();

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





/*
Perform "get line list" request
================================================================================================
*/ 

HRESULT TagDriverC::GetLineList (VirtualLineDesc * const Buffer, UINT const Size, UINT & NumLines) const {

  Assert (!IsBadWritePtr (Buffer, Size));

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Driver connection is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    DriverReq Req;

    Debug (6, L"Requesting line list");

    LocSize RetSize;

    Res = ControlRequest (IoCtl_Driver_GetLineList, Req, sizeof (Req), Buffer, Size, &RetSize);

    if (FAILED (Res)) {

      Debug (9, L"Cannot get line list (%s)", ErrorText (Res));

      goto Exit;

    }

    NumLines = RetSize / sizeof (*Buffer);

    Assert (NumLines * sizeof (*Buffer) == RetSize);

    Debug (7, L"Got %u line descriptors to %#x:%u", NumLines, Buffer, RetSize);

    Assert (Res == S_OK);

  }

Exit:

  Check ();

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





HRESULT TagDriverC::CreateLine (VirtualLineDesc const & Desc) const {

  Assert (!IsBadReadPtr (&Desc, sizeof (Desc)));
  Assert (!IsBadStringPtr (Desc.KsName, NumArrayElems (Desc.KsName)));
  Assert (!IsBadStringPtr (Desc.EpName, NumArrayElems (Desc.EpName)));

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Driver connection is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    DriverReq_CreateLine Req;

    Req.LineDesc = Desc;

    Debug (
      6,
      L"Creating line %u (KS name \"%s\", EP name \"%s\")",
      Req.LineDesc.Id, Req.LineDesc.KsName, Req.LineDesc.EpName
    );

    Res = ControlRequest (IoCtl_Driver_CreateLine, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot create line %u (%s)", Desc.Id, ErrorText (Res));

      goto Exit;

    }

    Debug (
      7,
      L"Created line %u (KS name \"%s\", EP name \"%s\")",
      Req.LineDesc.Id, Req.LineDesc.KsName, Req.LineDesc.EpName
    );

    Assert (Res == S_OK);

  }

Exit:

  Check ();

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





HRESULT TagDriverC::DeleteLine (UINT const Id) const {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Driver connection is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    DriverReq_DeleteLine Req;

    Req.Id = Id;

    Debug (6, L"Deleting line %u", Req.Id);

    Res = ControlRequest (IoCtl_Driver_DeleteLine, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot delete line %u (%s)", Req.Id, ErrorText (Res));

      goto Exit;

    }

    Debug (7, L"Deleted line %u", Req.Id);

    Assert (Res == S_OK);

  }

Exit:

  Check ();

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





HRESULT TagDriverC::CreatePipe (TagPipe * & Pipe, UINT const LineId, bool const Capture) const {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  TagPipe * P = nullptr;

  {

    if (!IsFound ()) {

      Debug (9, L"Driver interface was not successfully found");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    P = new TagPipeC (this, LineId, Capture);

    if (P == nullptr) {

      Debug (9, L"No memory for %s pipe", P->GetDir ());

      Res = HRESULT_FROM_WIN32 (ERROR_NOT_ENOUGH_MEMORY);
      goto Exit;

    }

    Debug (7, L"Created TagPipe object %#x", P);

    Res = P->Open ();

    if (FAILED (Res)) {

      goto Exit;

    }

    Pipe = P;

    Assert (SUCCEEDED (Res));

  }

Exit:

  if (FAILED (Res)) {

    if (P != nullptr) {

      delete P;
      P = nullptr;

    }

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





#if _Debug

void TagDriverC::SetDebugPrintLevelOffset (int const Offset) {

  Dbg.SetPrintLevelOffset (Offset);

}

#endif // _Debug





HRESULT _LinkageDef _stdcall CreateDriver (TagDriver * & Obj, GUID const & ClassId) {

  return TagDriverC::Create (Obj, ClassId);

}





} // namespace
