#include "stdafx.h"





namespace ThinAudioGateway {





/*
################################################################################################

Constants
================================================================================================

################################################################################################
*/





static PCTSTR const KsStateNames [] = {L"Stop", L"Acquire", L"Pause", L"Run" };





/*
################################################################################################

TagHostC class implementation
================================================================================================

################################################################################################
*/





/*
Constructor
================================================================================================
*/ 

TagHostC::TagHostC (TagDriver const * const Drv, UINT const LineId, bool const Capt, CallbackInterface * const CB)

: Driver (Drv), LineId (LineId), Capture (Capt), Callback (CB), CbWrapper (nullptr, nullptr)

#if _Debug

  , Dbg (&::Dbg, 0, L"")

#endif // _Debug

{

  PreInit ();

  Debug (6, L"Constructed");

}





/*
Constructor
================================================================================================
*/ 

TagHostC::TagHostC (
  TagDriver const * const Drv, UINT const LineId, bool const Capt, CopyDataFunc * const CopyData, void * const Context
)

: Driver (Drv), LineId (LineId), Capture (Capt), Callback (&CbWrapper), CbWrapper (CopyData, Context)

#if _Debug

  , Dbg (&::Dbg, 0, L"")

#endif // _Debug

{

  PreInit ();

  Debug (6, L"Constructed");

}





/*
Pre-initializer
================================================================================================
*/

void TagHostC::PreInit () {

  Assert (LineId != 0);

#if _Debug

  Dbg.SetName (L"Host %u%c", LineId, IsCapt ()? L'c' : L'r');

#endif

  Assert (Callback != nullptr);

  Event = 0;
  Pipe = nullptr;
  Started = false;
  EventLoopActive = false;
  ExitRequest = false;

  Fmt.Invalidate ();

}

/*
Destructor
================================================================================================
*/

TagHostC::~TagHostC () {

  Debug (6, L"Destructing");

  Stop ();

  Debug (6, L"Destructed");

}





/*
Create object instance with a callback interface
================================================================================================
*/ 

HRESULT _stdcall TagHostC::Create (
  TagHost * & Obj, TagDriver const * const Drv, UINT const LineId, bool const Capture, CallbackInterface * const CB
) {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  TagHostC * Host = nullptr;

  Obj = nullptr;

  {

#if _Debug

    GblDebug (7, L"Creating host object");

#endif _Debug

    Host = new TagHostC (Drv, LineId, Capture, CB);

    if (Host == nullptr) {

#if _Debug

      GblDebug (9, L"No memory for host object");

#endif _Debug

      Res = HRESULT_FROM_WIN32 (ERROR_NOT_ENOUGH_MEMORY);
      goto Exit;

    }

#if _Debug

    GblDebug (8, L"Created host object %#x", Host);

#endif _Debug

    Host->Check ();

    Obj = Host;

    Res = S_OK;

  }

Exit:

  if (FAILED (Res)) {

    if (Host != nullptr) {

      delete Host;
      Host = nullptr;

    }

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





/*
Create object instance with a callback function
================================================================================================
*/ 

HRESULT _stdcall TagHostC::Create (
  TagHost * & Obj, TagDriver const * const Drv, UINT const LineId, bool const Capture,
  CopyDataFunc * const CopyData, void * const Context
) {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  TagHostC * Host = nullptr;

  Obj = nullptr;

  {

#if _Debug

    GblDebug (7, L"Creating host object");

#endif _Debug

    Host = new TagHostC (Drv, LineId, Capture, CopyData, Context);

    if (Host == nullptr) {

#if _Debug

      GblDebug (9, L"No memory for host object");

#endif _Debug

      Res = HRESULT_FROM_WIN32 (ERROR_NOT_ENOUGH_MEMORY);
      goto Exit;

    }

#if _Debug

    GblDebug (8, L"Created host object %#x", Host);

#endif _Debug

    Host->Check ();

    Obj = Host;

    Res = S_OK;

  }

Exit:

  if (FAILED (Res)) {

    if (Host != nullptr) {

      delete Host;
      Host = nullptr;

    }

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





/*
Delete object instance
================================================================================================
*/ 

void TagHostC::Delete () {

  Debug (6, L"Deleting host object %#x", this);

  delete this;

}





/*
Start streaming
================================================================================================
*/

HRESULT TagHostC::Start (
  FormatRangeDesc const * const FormatRange, WaveFormatDesc const * const DefFormat, UINT const MinBufMcs
) {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (Event != 0 || Pipe != nullptr || Started) {

      Debug (9, _UT (__FUNCTION__) L" is called in the wrong state");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    Assert (!EventLoopActive && !ExitRequest);

    Res = Driver->CreatePipe (Pipe, GetLineId (), IsCapt ());

    if (FAILED (Res)) {

      Debug (9, L"Cannot obtain pipe (%s)", ErrorText (Res));

      goto Exit;

    }

    if (FormatRange != nullptr) {

      Res = Pipe->SetFormatRange (*FormatRange);

      if (FAILED (Res)) {

        Debug (9, L"Cannot set format range to %s (%s)", WfrText (*FormatRange), ErrorText (Res));

        goto Exit;

      }

      Debug (7, L"Format range set to %s", WfrText (*FormatRange));

    }

    if (DefFormat != nullptr) {

      Res = Pipe->SetDefaultFormat (*DefFormat);

      if (FAILED (Res)) {

        Debug (9, L"Cannot set default format range to %s (%s)", WfxText (*DefFormat), ErrorText (Res));

        goto Exit;

      }

      Debug (7, L"Default format set to %s", WfxText (*DefFormat));

    }

    if (MinBufMcs != 0) {

      // If the driver increases buffer size above requested by the Windows Audio Graph
      // (AudioDG), the Graph usually fails its client's request.

      Res = Pipe->SetMinimumBufferLength (MinBufMcs);

      if (FAILED (Res)) {

        Debug (9, L"Cannot set minimum buffer length to %u mcs (%s)", MinBufMcs, ErrorText (Res));

        goto Exit;

      }

    }

    Event = CreateEvent (nullptr, false, false, nullptr);

    if (Event == 0) {

      Res = HRESULT_FROM_WIN32 (GetLastError ());

      Debug (9, L"Cannot create notification event (%s)", ErrorText (Res));

      goto Exit;

    }

    Res = Pipe->SetNotificationEvent (Event);

    if (FAILED (Res)) {

      Debug (9, L"Cannot set notification event (%s)", ErrorText (Res));

      goto Exit;

    }

    Started = true;

    Debug (7, L"Started");

    Res = S_OK;

  }

Exit:

  if (FAILED (Res)) {

    Stop ();

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





/*
Stop streaming
================================================================================================
*/

HRESULT TagHostC::Stop () {

  if (EventLoopActive) {

    ExitRequest = true;

    while (EventLoopActive) {

      Sleep (15);

    }

    Assert (!ExitRequest);

  }

  Started = false;

  if (Pipe != nullptr) {

    Debug (6, L"Deleting pipe %#x", Pipe);

    Pipe->Delete ();
    Pipe = nullptr;

  }

  if (Event != 0) {

    Debug (6, L"Deleting event %#x", Event);

    CloseHandle (Event);
    Event = 0;

  }

  return S_OK;

}





/*
Event loop
================================================================================================
*/

HRESULT TagHostC::EventLoop (UINT const PollPeriodMs) {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (Pipe == nullptr || !Pipe->IsOpen () || EventLoopActive) {

      Debug (9, _UT (__FUNCTION__) L" is called in the wrong state");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    if (PollPeriodMs == 0) {

      Debug (9, L"Poll period cannot be 0");

      Res = E_INVALIDARG;
      goto Exit;

    }

    EventLoopActive = true;

    CommonDataDesc & CD = Pipe->GetCommonData ();
    StreamInfo & SI = CD.Drv.SI;

    PBYTE & Buffer = (PBYTE &) (CD.Host.Buffer);

    UINT BufferFrames = 0;
    UINT FramesInPacket = 0;

    bool HaveStream = false;

    KSSTATE State = KSSTATE_STOP;

    UINT & DevicePos = CD.Host.DevicePos;

    timeBeginPeriod (1);

    LARGE_INTEGER QpcFreq;

    Verify (QueryPerformanceFrequency (&QpcFreq));

    Debug (7, L"QPC frequency: %u", QpcFreq.QuadPart);

    Assert (QpcFreq.QuadPart != 0);

    LARGE_INTEGER LastTimeQpc;

    LastTimeQpc.QuadPart = 0;

    Debug (7, L"Starting event loop");



    for (;;) {

      if (ExitRequest) {

        Debug (6, L"Exit request detected");

        ExitRequest = false;

        Res = S_OK;
        break;

      }

      // Wait for a driver notification or for a timeout.

      DWORD const WaitRes = WaitForSingleObject (Event, PollPeriodMs);

      //Debug (5, L"WR=%u, TRE=%u", WaitRes, ThreadExitRequest);

      if (WaitRes == WAIT_OBJECT_0) {

        //Debug (
        //  7,
        //  L"FI.Flags: %#x, FI.HasStream: %u, SI.State: %u, SI.BufferFrames: %u",
        //  CD.Drv.FI.Flags, CD.Drv.FI.HasStream, SI.State, SI.BufferFrames
        //);

        if (TestBits (CD.Drv.FI.Flags, FilterFlags::WaitingForHost)) {

          // Detect stream creation/termination

          if (HaveStream != CD.Drv.FI.HasStream) {

            HaveStream = CD.Drv.FI.HasStream;

            if (HaveStream) {

              Debug (7, L"Stream created");

              Assert (State == KSSTATE_STOP);
              Assert (!Fmt.IsValid ());

              Res = Fmt.Init (const_cast <WAVEFORMATEXTENSIBLE &> (CD.Drv.SI.Fmt));

              Assert (Res == S_OK);

            } else {

              Debug (7, L"Stream terminated");

              Assert (DevicePos == 0);

              State = KSSTATE_STOP;
              Fmt.Invalidate ();

            }

          }
          


          // Detect buffer creation/deletion

          if (BufferFrames != SI.BufferFrames) {
          
            if (BufferFrames == 0 && SI.BufferFrames != 0) {
            
              Debug (7, L"Buffer acquired (%u frames)", SI.BufferFrames);

            } else if (BufferFrames != 0 && SI.BufferFrames == 0) {
            
              Debug (7, L"Buffer released");

              FramesInPacket = 0;

            } else {

              Debug (6, L"Buffer size changed (%u -> %u frames)", BufferFrames, SI.BufferFrames);

            }

            BufferFrames = SI.BufferFrames;

            Assert (BufferFrames % 2 == 0);

            FramesInPacket = BufferFrames / 2;

          }



          // Detect stream state changement

          if (State != SI.State) {

            Assert (SI.State <= NumArrayElems (KsStateNames));

            if (abs (int (State - SI.State)) > 1) {

              Debug (9, L"Lost some state transitions");

            }

            Debug (6, L"State transition %s -> %s", KsStateNames [State], KsStateNames [SI.State]);

            Debug (5, L"DevPos: %u", DevicePos);

            if (SI.State == KSSTATE_STOP) {

              Assert (State == KSSTATE_ACQUIRE);

              Debug (7, L"Stream stopped");

            } else if (SI.State == KSSTATE_RUN) {

              Assert (State == KSSTATE_PAUSE);

              Debug (7, L"Stream started");

              Verify (QueryPerformanceCounter (&LastTimeQpc));
              Assert (LastTimeQpc.QuadPart != 0);

            } else if (State == KSSTATE_RUN && SI.State == KSSTATE_PAUSE) {

              Debug (7, L"Stream paused");

            }

            State = SI.State;

          }

          // Confirm the driver request

          Res = Pipe->Notify (STATUS_SUCCESS);

          if (FAILED (Res)) {

            break;

          }

        } // Filter/stream state analyzing

        continue; // Nothing to do more after driver's notification

      } // Successful wait actions



      if (State != KSSTATE_RUN) {

        continue; // Wait for a next event

      }



      Assert (LastTimeQpc.QuadPart != 0);

      LARGE_INTEGER CurrentTimeQpc;

      Verify (QueryPerformanceCounter (&CurrentTimeQpc));

      Assert (CurrentTimeQpc.QuadPart != 0);

      UINT64 const ElapsedTime = (
        static_cast <UINT64> (CurrentTimeQpc.QuadPart) - static_cast <UINT64> (LastTimeQpc.QuadPart)
      );

      LastTimeQpc.QuadPart = CurrentTimeQpc.QuadPart;

#if _Debug

      UINT const Mcs = static_cast <UINT> ((ElapsedTime * 1000000 + QpcFreq.QuadPart / 2) / QpcFreq.QuadPart);

#endif

      UINT Frames = static_cast <UINT> (
        (ElapsedTime * Fmt.GetSamplingRate () + QpcFreq.QuadPart / 2) / QpcFreq.QuadPart
      );

      Debug (5, L"Period: %u mcs, %u fms", Mcs, Frames);

      Frames = min (Frames, FramesInPacket);

      // Copy data by portions

      while (Frames != 0) {

        Assert (DevicePos < BufferFrames);

        UINT const Portion = min (Frames, BufferFrames - DevicePos);

        Assert (Portion != 0);

        LocSize const Bytes = Fmt.FramesToBytes (Portion);

        LocSize const Actual = Callback->CopyData (Buffer + (DevicePos * Fmt.GetFrameSize ()), Bytes);

        Assert (Actual % Fmt.GetFrameSize () == 0);

        Debug (5, L"%s %u fms at %u", IsCapt ()? L"Written" : L"Read", Actual, DevicePos);

        Assert (DevicePos + Portion <= BufferFrames);

        Pipe->AdvanceDevicePosition (Fmt.BytesToFrames (Actual));

        Assert (Portion <= Frames);

        Frames -= Portion;

      } // Portion read/write loop



    } // Event loop


    Debug (7, L"Event loop ended");

    timeEndPeriod (1);

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  EventLoopActive = false;

  return Res;

}





#if _Debug

void TagHostC::SetDebugPrintLevelOffset (int const Offset) {

  Dbg.SetPrintLevelOffset (Offset);

}

#endif // _Debug





HRESULT _LinkageDef _stdcall CreateHost (
  TagHost * & Obj, TagDriver const * const Drv, UINT const LineId, bool const Capture, CallbackInterface * const CB
) {

  return TagHostC::Create (Obj, static_cast <TagDriver const *> (Drv), LineId, Capture, CB);

}





} // namespace
