#include "stdafx.h"





namespace ThinAudioGateway {





/*
################################################################################################

Constants
================================================================================================

################################################################################################
*/





/*
################################################################################################

Variables
================================================================================================

################################################################################################
*/





/*
################################################################################################

TagPipeC class implementation
================================================================================================

################################################################################################
*/





/*
Constructor
================================================================================================
*/ 

TagPipeC::TagPipeC (TagDriver const * xDrv, UINT const xLineId, bool const xCapt)

: Drv (xDrv), LineId (xLineId), Capture (xCapt)

#if _Debug

  , Dbg (&::Dbg, 0, L"Pipe %u%c", LineId, IsCapt ()? L'c' : L'r')

#endif

{

  Connection = 0;
  FrameSize = 0;
  CommonData = nullptr;

  Debug (6, L"Constructed");

  Check ();

}





/*
Destructor
================================================================================================
*/ 

TagPipeC::~TagPipeC () {

  Debug (6, L"Destructing");

  Check ();

  Close ();

}





/*
Delete pipe object
================================================================================================
*/ 

void TagPipeC::Delete () {

  Check ();

  Debug (6, L"Deleting TagPipe object %#x", this);

  delete this;

}





/*
Open (connect to the driver)
================================================================================================
*/ 

HRESULT TagPipeC::Open () {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (IsOpen ()) {

      Debug (9, L"Pipe is already open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    Assert (Drv->IsFound () && *Drv->GetInterface () != L'\0');

    wchar_t PipeInterface [MAX_PATH];

    _snwprintf (
      PipeInterface, NumArrayElems (PipeInterface),
      L"%s\\%s\\lines\\%u",
      Drv->GetInterface (), GuidText (Drv->GetClassId ()), GetLineId ()
    );

    Debug (6, L"Opening pipe connection: %s", PipeInterface);

    Connection = CreateFile (
      PipeInterface, GENERIC_READ | GENERIC_WRITE, 0, nullptr, OPEN_EXISTING, 0, 0
    );

    if (Connection == INVALID_HANDLE_VALUE) {

      Connection = 0;

      Res = HRESULT_FROM_WIN32 (GetLastError ());

      if (Res == HRESULT_FROM_WIN32 (ERROR_FILE_NOT_FOUND)) {

        Debug (9, L"Pipe interface %s not found", PipeInterface);

      } else {

        Debug (9, L"Cannot open pipe interface %s (%s)", PipeInterface, ErrorText (Res));

      }

      goto Exit;

    }

    Debug (7, L"Pipe connection is open (%#x)", Connection);

    Assert (CommonData == nullptr);

    LineReq_GetCommonDataAddress Req;

    Res = ControlRequest (
      IoCtl_Line_GetCommonDataAddress, Req, sizeof (Req), &CommonData, sizeof (CommonData)
    );

    if (FAILED (Res)) {

      Debug (9, L"Cannot get common data address (%s)", ErrorText (Res));

      goto Exit;

    }

    Debug (7, L"Common data: %Ix:%u", CommonData, sizeof (*CommonData));

    if (IsBadReadPtr (CommonData, sizeof (*CommonData))) {

      Debug (9, L"Common data block (%Ix:%u) is not readable", CommonData, sizeof (*CommonData));

      Res = HRESULT_FROM_WIN32 (ERROR_ACCESS_DENIED);
      goto Exit;

    }

    if (!!TestBits (CommonData->Drv.FI.Flags, FilterFlags::Capture) != IsCapt ()) {

      Debug (9, L"Line %u direction doesn't match pipe direction", GetLineId ());

      Res = HRESULT_FROM_WIN32 (ERROR_NO_MATCH);
      goto Exit;

    }

    Debug (
      7,
      L"Audio buffer: %Ix:%u",
      CommonData->Host.Buffer, CommonData->Drv.FI.ReservedBufferSize
    );

    if (IsBadReadPtr (CommonData->Host.Buffer, CommonData->Drv.FI.ReservedBufferSize)) {

      Debug (
        9,
        L"Audio buffer (%Ix:%u) is not readable",
        CommonData->Host.Buffer, CommonData->Drv.FI.ReservedBufferSize
      );

      Res = HRESULT_FROM_WIN32 (ERROR_ACCESS_DENIED);
      goto Exit;

    }

    Res = S_OK;

  }

Exit:

  if (FAILED (Res)) {

    Close ();

  }

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





/*
Close (disconnect from the driver)
================================================================================================
*/ 

void TagPipeC::Close () {

  Check ();

  if (IsOpen ()) {

    CommonData = nullptr;

    Debug (6, L"Closing connection %#x", Connection);

    CloseHandle (Connection);

    Debug (7, L"Connection %#x closed", Connection);

    Connection = 0;

  }

  Check ();

}





/*
Issue a control request
================================================================================================
*/ 

HRESULT TagPipeC::ControlRequest (
  UINT const Func,
  DriverReq & Req, LocSize const ReqSize,
  void * const Data, LocSize const DataSize,
  LocSize * const RetSize
) const {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Pipe is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    // Bypass checking if not fully initialized

    if (Func != IoCtl_Line_GetCommonDataAddress) {

      Check ();

    }

    Res = IoControlRequest (Connection, Func, Req, ReqSize, Data, DataSize, RetSize);

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





/*
Set minimum buffer size for a client
================================================================================================
*/ 

HRESULT TagPipeC::SetMinimumBufferLength (UINT const Mcs) const {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Pipe is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    LineReq_SetMinimumBufferLength Req;

    Req.Mcs = Mcs;

    Debug (6, L"Setting minimum buffer duration to %u mcs", Req.Mcs);

    Res = ControlRequest (IoCtl_Line_SetMinimumBufferLength, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot set buffer duration to %u mcs (%s)", Req.Mcs, ErrorText (Res));

      goto Exit;

    }

    Debug (7, L"Minimum buffer duration set to %u mcs", Req.Mcs);

    Res = S_OK;

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





/*
Set audio format available for driver's subdevice clients
================================================================================================
*/ 

HRESULT TagPipeC::SetFormatRange (FormatRangeDesc const & FRD) const {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Pipe is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    //WaveFormatRange const & FR = FR.from

    //Assert (Range.IsCorrect ());

    LineReq_SetFormatRange Req;

    Req.Range = FRD;

    Debug (6, L"Setting filter format range to %s", WfrText (Req.Range));

    Res = ControlRequest (IoCtl_Line_SetFormatRange, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot set filter format range to %s (%s)", WfrText (Req.Range), ErrorText (Res));

      goto Exit;

    }

    Debug (7, L"Filter format range set to %s", WfrText (Req.Range));

    Res = S_OK;

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





/*
Set default audio format for the driver's subdevice
================================================================================================
*/ 

HRESULT TagPipeC::SetDefaultFormat (WaveFormatDesc const & FD) const {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Pipe is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    LineReq_SetDefaultFormat Req;

    Req.Format = FD;

    Debug (6, L"Setting default filter format to %s", WfxText (Req.Format));

    Res = ControlRequest (IoCtl_Line_SetDefaultFormat, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot set default filter format to %s (%s)", WfxText (Req.Format), ErrorText (Res));

      goto Exit;

    }

    Debug (7, L"Default filter format set to %s", WfxText (Req.Format));

    Res = S_OK;

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





/*
Set a notification event
================================================================================================
*/ 

HRESULT TagPipeC::SetNotificationEvent (HANDLE const Event) const {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Pipe is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    LineReq_SetNotificationEvent Req;

    Req.Event = Event;

    Debug (6, L"Setting notification event to %#x", Req.Event);

    Res = ControlRequest (IoCtl_Line_SetNotificationEvent, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot set notification event to %#x (%s)", Req.Event, ErrorText (Res));

      goto Exit;

    }

    Debug (7, L"Notification event %#x passed to the filter", Event);

    Assert (Res == S_OK);

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





/*
Notify driver about host's state change
================================================================================================
*/

HRESULT TagPipeC::Notify (NTSTATUS const Code) {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Pipe is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    Debug (5, L"Notifying driver (%#x)", Code);

    LineReq_NotifyDriver Req;

    Req.Code = Code;

    Res = ControlRequest (IoCtl_Line_Notify, Req, sizeof (Req));

    if (FAILED (Res)) {

      Debug (9, L"Cannot notify driver with code %#x (%s)", ErrorText (Res));

      goto Exit;

    }

    Assert (Res == S_OK);

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





/*
Reset host/device audio position (read/write buffer offset)
================================================================================================
*/
// On entering KSSTATE_STOP, position is reset by the filter.
//

HRESULT TagPipeC::ResetDevicePosition () {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Pipe is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    CommonData->Host.DevicePos = 0;
    CommonData->Host.DevicePosBytes = 0;

    Debug (6, L"Device position reset to 0");

    Check ();

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return S_OK;

}





/*
Advance host/device audio position (read/write buffer offset)
================================================================================================
*/ 

HRESULT TagPipeC::AdvanceDevicePosition (UINT const Frames) {

  Check ();

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    if (!IsOpen ()) {

      Debug (9, L"Pipe is not open");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    if (Frames == 0) {

      Debug (9, L"Number of frames to advance is 0");

      Res = E_UNEXPECTED;
      goto Exit;

    }

    UINT & BufferFrames = CommonData->Drv.SI.BufferFrames;
    UINT & DevPos = CommonData->Host.DevicePos;

    Assert (BufferFrames != 0);

#if _Debug

    UINT const PrevPos = DevPos;

#endif

    // Update main position in frames

    DevPos = (DevPos + Frames) % BufferFrames; // Must be written once

    // Update auxiliary position in bytes (used as a position register for the client)

    WaveFormat const & Fmt = Fmt.FromWfx (const_cast <WAVEFORMATEXTENSIBLE const &> (CommonData->Drv.SI.Fmt));

    CommonData->Host.DevicePosBytes = Fmt.FramesToBytes (DevPos); // Must be written once

    Debug (5, L"Adv dev pos by %u fms (%u->%u)", Frames, PrevPos, DevPos);

    Res = Notify (StatusCodes::CommonDataUpdated);

    if (FAILED (Res)) {

      goto Exit;

    }

    Assert (Res == S_OK);

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  Check ();

  return Res;

}





#if _Debug

void TagPipeC::SetDebugPrintLevelOffset (int const Offset) {

  Dbg.SetPrintLevelOffset (Offset);

}

#endif // _Debug





} // namespace
