#include "stdafx.h"





namespace ThinAudioGateway {





/*
################################################################################################

Variables
================================================================================================

################################################################################################
*/





/*
################################################################################################

Implementation of ErrorText_i
================================================================================================

################################################################################################
*/





LocSize ErrorText_i::ToStr (PTSTR const Buf, UINT const MaxLen, DWORD const Code) {

  DWORD const LastErr = GetLastError ();

  PTSTR cp = Buf;
  UINT Avail = MaxLen;

  _snwprintf (
    cp, Avail,
    L"%s%u/%u - ",
    (Code & APPLICATION_ERROR_MASK)? L"u" : L"",
    static_cast <UINT> (HRESULT_FACILITY (Code)), static_cast <UINT> (HRESULT_CODE (Code))
  );

  cp [Avail - 1] = L'\0';

  UINT const Offset = static_cast <UINT> (wcslen (Buf));

  cp += Offset;
  Avail -= Offset;

  UINT Len;

  Len = FormatMessage (
    FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK,
    nullptr,
    Code,
    MAKELANGID (LANG_NEUTRAL, SUBLANG_DEFAULT),
    cp,
    Avail,
    nullptr
  );

  if (Len != 0) {

    cp += Len - 1;

    if (*cp == L' ') {

      *cp = L'\0';

    }

  } else if (cp > Buf + 3 && cp [-1] == L' ' && cp [-2] == L'-' && cp [-3] == L' ') {

    cp [-3] = L'\0';             // Cut " - "

  }

  Assert (wcslen (Buf) < MaxLen);

  SetLastError (LastErr);

  return static_cast <LocSize> (wcslen (Buf));

}





/*
################################################################################################

Implementation of GuidText_i
================================================================================================

################################################################################################
*/





LocSize GuidText_i::ToStr (PTSTR const Buffer, UINT const MaxLen, GUID const & Guid) {

  wchar_t Str [40];

  if (&Guid) {

    _snwprintf (
      Str, NumArrayElems (Str),
      L"%08lx-%04x-%04x-%02x%02x-%02x%02x%02x%02x%02x%02x",
      Guid.Data1,
      Guid.Data2, Guid.Data3,
      Guid.Data4 [0], Guid.Data4 [1],
      Guid.Data4 [2], Guid.Data4 [3], Guid.Data4 [4],
      Guid.Data4 [5], Guid.Data4 [6], Guid.Data4 [7]
    );

  } else {

    wcscpy (Str, L"GUID_NULL");

  }

  UINT const Len = static_cast <UINT> (wcslen (Str));
  UINT const ActualLen = min (Len, MaxLen - 1);

  memcpy (Buffer, Str, StrLenToSz (ActualLen));

  Buffer [ActualLen] = L'\0';

  return ActualLen;

}





/*
################################################################################################

Implementation of WaveFormat
================================================================================================

################################################################################################
*/





//
// Get format subtype ID (WAVE_FORMAT_xxx)
// =============================================================================================
//

UINT WaveFormat::GetSubType () const {

  UINT const Tag = GetTag ();

  return (Tag == WAVE_FORMAT_EXTENSIBLE)? EXTRACT_WAVEFORMATEX_ID (&Fmt.SubFormat) : Tag;

}





//
// Get subformat GUID
// =============================================================================================
//

GUID const WaveFormat::GetSubFormat () const {

  if (IsExtensible ()) {

    return Fmt.SubFormat;

  } else {

    GUID SubFormat;

    INIT_WAVEFORMATEX_GUID (&SubFormat, GetTag ());

    return SubFormat;

  }

}





bool WaveFormat::IsExtensible () const {

  return (
    GetTag () != WAVE_FORMAT_PCM
    && Fmt.Format.cbSize >= sizeof (Fmt) - sizeof (Fmt.Format)
  );

}





bool WaveFormat::IsPcm () const {

  return (
    GetSubFormat () == KSDATAFORMAT_SUBTYPE_PCM
    || GetSubFormat () == KSDATAFORMAT_SUBTYPE_IEEE_FLOAT
  );

}





bool WaveFormat::IsWfx () const {

  return (GetTag () == WAVE_FORMAT_PCM || Fmt.Format.cbSize == 0);

}





bool WaveFormat::IsCorrect () const {

  Assume (sizeof (WaveFormat) == sizeof (WAVEFORMATEXTENSIBLE));

  bool Success = false;

  {

    if (!IsValid ()) {

      Debug (9, L"Format is not valid");

      goto Exit;

    }

    if (!IsInRange (Fmt.Format.nSamplesPerSec, SamplingRateMin, SamplingRateMax)) {

      Debug (9, L"Incorrect nSamplesPerSec %u", Fmt.Format.nSamplesPerSec);

      goto Exit;

    }

    if (!IsInRange (Fmt.Format.wBitsPerSample, BitsPerSampleMin, BitsPerSampleMax)) {

      Debug (9, L"Incorrect wBitsPerSample%u", Fmt.Format.wBitsPerSample);

      goto Exit;

    }

    if (Fmt.Format.wBitsPerSample % 8 != 0) {

      Debug (9, L"wBitsPerSample (%u) is not a multiple of 8", GetBitsPerSample ());

      goto Exit;

    }

    if (!IsInRange (Fmt.Format.nChannels, NumChannelsMin, NumChannelsMax)) {

      Debug (9, L"Incorrect nChannels %u", Fmt.Format.nChannels);

      goto Exit;

    }

    if (IsUniform ()) {

      if (GetAlign () != GetNumChannels () * GetContainerSize ()) {

        Debug (
          9,
          L"nBlockAlign (%u) != nChannels (%u) * wBitsPerSample (%u) / 8",
          GetAlign (), GetNumChannels (), GetBitsPerSample ()
        );

        goto Exit;

      }

      if (Fmt.Format.nAvgBytesPerSec != GetAlign () * GetSamplingRate ()) {

        Debug (
          9,
          L"nAvgBytesPerSec (%u) != nBlockAlign (%u) * nSamplesPerSec (%u)",
          Fmt.Format.nAvgBytesPerSec, GetAlign (), GetSamplingRate ()
        );

        goto Exit;

      }

    }

    if (IsExtensible ()) {

      if (GetTag () == WAVE_FORMAT_PCM) {

        Debug (9, L"WAVE_FORMAT_PCM in the extensible format");

        goto Exit;

      }

      if (GetTag () != WAVE_FORMAT_EXTENSIBLE) {

        Debug (9, L"Invalid tag %#x in the extensible format");

        goto Exit;

      }

      if (!IS_VALID_WAVEFORMATEX_GUID (&Fmt.SubFormat)) {

        Debug (9, L"Invalid subformat GUID %s", GuidText (Fmt.SubFormat));

        goto Exit;

      }

      if (!IsInRange (GetActualBitsPerSample (), 1, Fmt.Format.wBitsPerSample)) {

        Debug (
          9,
          L"wValidBitsPerSample (%u) is out of %u..%u",
          GetActualBitsPerSample (), 1, Fmt.Format.wBitsPerSample
        );

        goto Exit;

      }

    } // IsExtensible

    Success = true;

  }

Exit:

  return Success;

}





HRESULT WaveFormat::Init (UINT const SR, UINT const BPS, UINT const NC, UINT const BPC, UINT const CC) {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    Fmt.Format.wFormatTag = WAVE_FORMAT_UNKNOWN;

    Debug (
      7,
      L"Initializing format with SR %u, BPS %u, NC %u, BPC %u, CC %#x",
      SR, BPS, NC, BPC, CC
    );

    if (!IsInRange (SR, SamplingRateMin, SamplingRateMax)) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"SR (%u) is out of %u..%u", SR, SamplingRateMin, SamplingRateMax);

      goto Exit;

    }

    if (!IsInRange (BPS, BitsPerSampleMin, BitsPerSampleMax)) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"BPS (%u) is out of %u..%u", BPS, BitsPerSampleMin, BitsPerSampleMax);

      goto Exit;

    }

    if (!IsInRange (NC, NumChannelsMin, NumChannelsMax)) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"NC (%u) is out of %u..%u", NC, NumChannelsMin, NumChannelsMax);

      goto Exit;

    }

    if (BPC > 4) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"BPC (%u) is out of 0..4", BPC);

      goto Exit;

    }

    if (BPC != 0 && BPC < (BPS + 7) / 8) {

      Res = HRESULT_FROM_WIN32 (ERROR_BAD_FORMAT);

      Debug (9, L"BPC (%u) cannot hold %u bits", BPC, BPS);

      goto Exit;

    }

    Fmt.Format.nSamplesPerSec = SR;
    Fmt.Format.nChannels = static_cast <WORD> (NC);
    Fmt.Format.nBlockAlign = static_cast <WORD> (((BPC != 0)? BPC : (BPS + 7 / 8)) * NC);

    if (
      (BPS == 8 || BPS == 16)
      && NC <= 2
      && BPS == GetContainerSize () * 8
      && (CC == UINT_MAX || CC == GetStdChannelMask (NC))
    ) {

      Fmt.Format.wFormatTag = WAVE_FORMAT_PCM;
      Fmt.Format.wBitsPerSample = static_cast <WORD> (BPS);
      Fmt.Format.cbSize = 0;

    } else {

      Fmt.Format.wFormatTag = WAVE_FORMAT_EXTENSIBLE;

      Fmt.Format.wBitsPerSample = static_cast <WORD> (TestBits (BPS + 7, ~0x7));

      Fmt.SubFormat = KSDATAFORMAT_SUBTYPE_PCM; // Use INIT_WAVEFORMATEX_GUID for non-PCM
      Fmt.Format.cbSize = sizeof (Fmt) - sizeof (Fmt.Format);
      Fmt.Samples.wValidBitsPerSample = static_cast <WORD> (GetBitsPerSample ());
      Fmt.dwChannelMask = (CC != UINT_MAX)? CC : GetStdChannelMask (NC);

    }

    Debug (7, L"WF initialized: %s", WfxText (*this));

    Check ();

    Res = S_OK;

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





HRESULT WaveFormat::Init (WAVEFORMATEXTENSIBLE const & Wfx) {

  Fmt = Wfx;

  HRESULT FinalRes;

  if (IsCorrect ()) {

    FinalRes = S_OK;

  } else {

    Invalidate ();

    FinalRes = E_INVALIDARG;

  }

  return FinalRes;

}





UINT WaveFormat::GetStdChannelMask (UINT const NumChans) {

  Assert (NumChans >= NumChannelsMin && NumChans <= NumChannelsMax);

  static const UINT MaskTable [] = {

    KSAUDIO_SPEAKER_MONO,       // 1
    KSAUDIO_SPEAKER_STEREO,     // 2
    KSAUDIO_SPEAKER_DIRECTOUT,  // 3
    KSAUDIO_SPEAKER_QUAD,       // 4
    KSAUDIO_SPEAKER_DIRECTOUT,  // 5
    KSAUDIO_SPEAKER_5POINT1_SURROUND, // 6
    KSAUDIO_SPEAKER_DIRECTOUT,  // 7
    KSAUDIO_SPEAKER_7POINT1_SURROUND, // 8

  };

  return (NumChans <= NumArrayElems (MaskTable))? (
    MaskTable [NumChans - 1]
  ) : (
    KSAUDIO_SPEAKER_DIRECTOUT
  );

}





void WaveFormat::ToWfd (WaveFormatDesc & FD) const {

  Assert (IsCorrect ());

  FD.SamplingRate = GetSamplingRate ();
  FD.BitsPerSample = GetActualBitsPerSample ();
  FD.NumChannels = GetNumChannels ();
  FD.BytesPerContainer = GetContainerSize ();
  FD.ChannelConfig = GetChannelConfig ();

}





LocSize WaveFormat::ToStr (PTSTR const Buffer, UINT const MaxLen) const {

  int const IR = _snwprintf (
    Buffer, MaxLen,
    L"%u/%u(%u)/%u(%x)",
    GetSamplingRate (), GetBitsPerSample (), GetContainerSize (), GetNumChannels (), GetChannelConfig ()
  );

  return (IR > 0)? static_cast <LocSize> (IR) : 0;

}





/*
################################################################################################

Implementation of WaveFormatRange
================================================================================================

################################################################################################
*/





bool WaveFormatRange::IsCorrect () const {

  bool const Success = (
    SamplingRate.IsCorrect ()
    && BitsPerSample.IsCorrect ()
    && NumChannels.IsCorrect ()
    && BytesPerContainer.IsCorrect ()
  );

  return Success;

}





LocSize WaveFormatRange::ToStr (PTSTR const Buffer, LocSize const MaxLen) const {

  int const IR = _snwprintf (
    Buffer, MaxLen,
    L"%s/%s/%s/%s",
    ScalarRangeText (RangeBoundType, SamplingRate), ScalarRangeText (RangeBoundType, BitsPerSample),
    ScalarRangeText (RangeBoundType, NumChannels), ScalarRangeText (RangeBoundType, BytesPerContainer)
  );

  return (IR > 0)? static_cast <LocSize> (IR) : 0;

}





/*
################################################################################################

Global functions
================================================================================================

################################################################################################
*/





/*
Issue a control request
================================================================================================
*/ 

HRESULT IoControlRequest (
  HANDLE const Connection, UINT const Func,
  DriverReq & Req, LocSize const ReqSize,
  void * const Data, LocSize const DataSize,
  LocSize * const RetSize
) {

  Assert (Connection != 0 && Func != 0);
  Assert (&Req != nullptr && ReqSize >= sizeof (DriverReq));

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    Req.ApiVer = ApiVersion;

    Debug (5, L"Driver req: API ver %u, func %#x, size %#x", Req.ApiVer, Func, ReqSize);

    DWORD BytesRet;

    bool const Success = !!DeviceIoControl (
      Connection, Func,
      &Req, static_cast <UINT> (ReqSize), Data, static_cast <UINT> (DataSize),
      &BytesRet, nullptr
    );

    if (RetSize != nullptr) {

      *RetSize = BytesRet;

    }

    if (!Success) {

      Res = HRESULT_FROM_WIN32 (GetLastError ());

      Debug (6, L"*** Driver request (%#x) failed (%s)", Func, ErrorText (Res));

      goto Exit;

    }

    Res = S_OK;

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





} // namespace
