#include "stdafx.h"





/*
################################################################################################

Constants
================================================================================================

################################################################################################
*/





static UINT const SamplingRateList [] = {

  5000, 8000, 11025, 22050, 44100, 48000, 88200, 96000, 192000, UINT_MAX

};

static UINT const BitsPerSampleList [] = {

  8, 16, 18, 20, 22, 24, 32, UINT_MAX

};

static UINT const NumChannelList [] = {

  1, 2, 4, 6, 8, UINT_MAX

};

static UINT const BytesPerContainerList [] = {

  0, 1, 2, 3, 4, UINT_MAX

};

static PCTSTR const ChannelConfigList [] = {

  L"def", L"0", L"1", L"3", L"4", L"3f", L"6f", L"63f", L"ff", nullptr

};

static PCTSTR const MinBufferMsList [] = {

  L"auto", L"1", L"2", L"5", L"10", L"20", L"50", nullptr

};



static UINT const FormatControls [] = {

  IDC_SamplingRate, IDC_BitsPerSample, IDC_NumChannels, IDC_BytesPerContainer,
  IDC_ChannelConfig,

  0

};

static UINT const NonRealtimeControls [] = {

  IDC_MinBufferMs, IDC_File, IDC_SelectFile,

  0

};

static PCTSTR const KsStateNames [] = {L"Stop", L"Acquire", L"Pause", L"Run" };





/*
################################################################################################

Implementation of Worker class
================================================================================================

################################################################################################
*/





/*
Constructor
================================================================================================
*/

Worker::Worker (Application const & xApp, UINT const xId, bool Capt, PCTSTR const xEpName)
:
  App (xApp), Id (xId), Capture (Capt),

  Host (nullptr),
  Dlg (0),
  Thread (0),
  ThreadExitRequest (false),
  Started (false),

  FileData (nullptr),
  FileSize (0),
  FilePos (0)

#if _Debug

  , Dbg (&::Dbg, 0, L"Wrk %u%c", Id, IsCapt ()? L'c' : L'r')

#endif // _Debug

{

  wcsncpy (EpName, xEpName, NumArrayElems (EpName));
  EpName [NumArrayElems (EpName) - 1] = L'\0';

  *FilePath = L'\0';

  Debug (6, L"Constructed");
  
}





/*
Destructor
================================================================================================
*/

Worker::~Worker () {

  if (Dlg != 0) {

    Debug (6, L"Destroying dialog %#x", Dlg);

    DestroyWindow (Dlg);
    Dlg = 0;

  }

  Debug (6, L"Destructed");
  
}





/*
Enable/disable non-realtime controls
================================================================================================
*/

void Worker::EnableNonRealtimeControls (bool const Enable) {

  if (!IsCapt ()) {

    EnableControls (Dlg, FormatControls, Enable);

  }

  EnableControls (Dlg, NonRealtimeControls, Enable);

}





/*
Read data from a PCM file
================================================================================================
*/

HRESULT Worker::ReadDataFromFile () {

  Assert (IsCapt ());

  Assert (FileData == nullptr && FileSize == 0 && FilePos == 0);

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    Assert (*FilePath == L'\0');

    UINT const PathLen = GetDlgItemText (Dlg, IDC_File, FilePath, NumArrayElems (FilePath));

    if (PathLen == 0) {

      Error (L"No file specified");

      goto Exit;

    }

    Res = ReadPcmFile (FilePath, Fmt, (PVOID &) (FileData), FileSize);

    if (FAILED (Res)) {

      Error (L"Cannot read PCM file %s (%s)", FilePath, ErrorText (Res));

      goto Exit;

    }

    Debug (7, L"%u PCM bytes read from %", FileSize, FilePath);

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





/*
Write data buffer to a PCM file
================================================================================================
*/

HRESULT Worker::WriteDataToFile () {

  Assert (!IsCapt ());

  Assert (FileData != nullptr && FileSize != 0 && FilePos < FileSize);

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  {

    Assert (*FilePath != L'\0');

    Debug (7, L"Writing %u PCM bytes to %s", FilePos, FilePath);

    Res = WritePcmFile (FilePath, Fmt, FileData, FilePos);

    if (FAILED (Res)) {

      Error (L"Cannot write %u bytes of PCM data to %s (%s)", FilePath, ErrorText (Res));

      goto Exit;

    }

  }

Exit:

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





/*
Copy data from/to host buffer
================================================================================================
*/

LocSize Worker::CopyData (void * const Addr, LocSize const Bytes) {

  Assert (Addr != nullptr && Bytes != 0);

  LocSize const Portion = min (Bytes, FileSize - FilePos);

  Assert (Portion != 0);

  if (IsCapt ()) {

    RtlCopyMemory (Addr, FileData + FilePos, Portion);

    Debug (5, L"Written %u file bytes at %u", Portion, FilePos);

  } else {

    RtlCopyMemory (FileData + FilePos, Addr, Portion);

    Debug (5, L"Read %u file bytes at %u", Portion, FilePos);

  }

#if _Debug

  LocSize const FilePosPrev = FilePos;

#endif

  FilePos += Portion;

  Assert (FilePos <= FileSize);

  if (FilePos == FileSize) {

    FilePos = 0; // Loop file buffer

  }

  Debug (5, L"File pos: %u->%u", FilePosPrev, FilePos);

  return Portion;

}





/*
C wrapper for CopyData 
================================================================================================
*/

LocSize Worker::CopyDataWrapper (void * const Context, void * const Addr, LocSize const Bytes) {

  Worker * const Wrk = static_cast <Worker *> (Context);

  Assert (Wrk != nullptr);

  return Wrk->CopyData (Addr, Bytes);

}





/*
Event loop
================================================================================================
*/

HRESULT Worker::EventLoop () {

  Assert (Host != nullptr);

#if UseLegacyWrappers

  Assert (Host_IsStarted (Host));

#else

  Assert (Host->IsStarted ());

#endif

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  Assert (Dlg != 0);

  Debug (7, L"Event loop started");

#if UseLegacyWrappers

  Res = Host_EventLoop (Host, 5);

#else

  Res = Host->EventLoop (5);

#endif

  Debug (7, L"Event loop ended");

  Assert (Res != E_I_PROCESSING_NOT_FINISHED);

  return Res;

}





/*
Thread function
================================================================================================
*/

HRESULT Worker::ThreadProc () {

  Debug (7, L"Thread started");

  HRESULT HR = E_I_PROCESSING_NOT_FINISHED;

  HANDLE MmcssTask = 0;

  {

    DWORD TaskIndex = 0;

    static wchar_t MmcssClass [] = L"Pro Audio";

    MmcssTask = AvSetMmThreadCharacteristics (MmcssClass, &TaskIndex);

    if (MmcssTask != 0) {

      Debug (7, L"Task registered in MMCSS \"%s\" class (%u)", MmcssClass, TaskIndex);

    } else {

      HR = HRESULT_FROM_WIN32 (GetLastError ());

      Debug (9, L"Cannot register task in MMCSS \"%s\" class (%s)", MmcssClass, ErrorText (HR));

      goto Exit;

    }

    HR = EventLoop ();

  }

Exit:

  Debug (7, L"Thread is exiting with %#x", HR);

  Assert (HR != E_I_PROCESSING_NOT_FINISHED);

  return HR;

}





/*
Thread function (outer)
================================================================================================
*/

DWORD WINAPI Worker::ThreadProcOuter (void * const Arg) {

  Assert (Arg != nullptr);

  Worker * const Wrk = reinterpret_cast <Worker *> (Arg);

  HRESULT const HR = Wrk->ThreadProc ();

  return static_cast <DWORD> (HR);

}





HRESULT Worker::SelectInputFile () {

  wchar_t TempPath [MAX_PATH];

  *TempPath = L'\0';

  OPENFILENAME OF;

  ZeroObject (OF);

  OF.lStructSize = sizeof (OF);
  OF.hwndOwner = Dlg;
  OF.hInstance = static_cast <HINSTANCE> (GetModuleHandle (nullptr));
  OF.lpstrInitialDir = nullptr;
  OF.lpstrFilter = L"Wave files (*.wav)\0*.wav\0";
  OF.lpstrFile = TempPath;
  OF.nMaxFile = NumArrayElems (TempPath);
  OF.lpstrTitle = L"Select a file to load";
  OF.lpstrDefExt = L"wav";
  OF.Flags = OFN_DONTADDTORECENT;

  HRESULT Res;

  if (GetOpenFileName (&OF)) {

    Debug (7, L"File load path: \"%s\"", TempPath);

    SetDlgItemText (Dlg, IDC_File, TempPath);

    Res = S_OK;

  } else {

    Debug (9, L"Input file selection failed (%#x)", CommDlgExtendedError ());

    Res = E_FAIL;

  }

  return Res;

}





HRESULT Worker::SelectOutputFile () {

  wchar_t TempPath [MAX_PATH];

  *TempPath = L'\0';

  OPENFILENAME OF;

  ZeroObject (OF);

  OF.lStructSize = sizeof (OF);
  OF.hwndOwner = Dlg;
  OF.hInstance = static_cast <HINSTANCE> (GetModuleHandle (nullptr));
  OF.lpstrInitialDir = nullptr;
  OF.lpstrFilter = L"Wave files (*.wav)\0*.wav\0";
  OF.lpstrFile = TempPath;
  OF.nMaxFile = NumArrayElems (TempPath);
  OF.lpstrTitle = L"Select a file to save data to";
  OF.lpstrDefExt = L"cfg";
  OF.Flags = OFN_DONTADDTORECENT | OFN_OVERWRITEPROMPT;

  HRESULT Res;

  if (GetSaveFileName (&OF)) {

    Debug (8, L"File save path: \"%s\"", TempPath);

    SetDlgItemText (Dlg, IDC_File, TempPath);

    Res = S_OK;

  } else {

    Debug (9, L"Output file selection failed (%x)", CommDlgExtendedError ());

    Res = E_FAIL;

  }

  return Res;

}





/*
Error message
================================================================================================
*/

void _cdecl Worker::Error (PCTSTR const Format, ...) const {

  va_list Args;

  va_start (Args, Format);
  ::Error (Dlg, Format, Args);
  va_end (Args);

}





void Worker::ResetControls () {

  EnableNonRealtimeControls (true);

  SetDlgItemText (Dlg, IDC_StartStop, L"Start");
  EnableWindow (GetDlgItem (Dlg, IDC_StartStop), true);

  SetDlgItemText (Dlg, IDC_WorkerState, L"Offline");

}





/*
Pipe dialog function
================================================================================================
*/

INT_PTR Worker::DialogProc (UINT const Msg, WPARAM const WP, LPARAM const LP) {

  Unreferenced (LP);

  Debug (5, L"Dlg: msg %#x, WP %#x, LP %#x", Msg, WP, LP);

  INT_PTR Res = 0;

  switch (Msg) {


    case WM_INITDIALOG: {

      Assert (Dlg != 0);

      Debug (6, L"Dlg init, win %#x", Dlg);

      wchar_t Text [128];

      wsprintf (Text, L"Pipe for %s line %u (%s)", GetDir (), GetId (), EpName);

      SetWindowText (Dlg, Text);

      Fmt.Init (SamplingRateDef, BitsPerSampleDef, NumChannelsDef);

      if (IsCapt ()) {

        EnableControls (Dlg, FormatControls, false);
        SetDlgItemText (Dlg, IDC_FileType, L"Load from");
        wsprintf (Text, L"playback_audio%u.wav", GetId ());

      } else {

        SetDlgItemText (Dlg, IDC_FileType, L"Save to");
        wsprintf (Text, L"recorded_audio%u.wav", GetId ());

      }

      SetDlgItemText (Dlg, IDC_File, Text);

      FillComboBox (Dlg, IDC_SamplingRate, SamplingRateList);
      FillComboBox (Dlg, IDC_BitsPerSample, BitsPerSampleList);
      FillComboBox (Dlg, IDC_NumChannels, NumChannelList);
      FillComboBox (Dlg, IDC_BytesPerContainer, BytesPerContainerList);
      FillComboBox (Dlg, IDC_ChannelConfig, ChannelConfigList);

      SetDlgItemInt (Dlg, IDC_SamplingRate, Fmt.GetSamplingRate (), false);
      SetDlgItemInt (Dlg, IDC_BitsPerSample, Fmt.GetBitsPerSample (), false);
      SetDlgItemInt (Dlg, IDC_NumChannels, Fmt.GetNumChannels (), false);
      SetDlgItemInt (Dlg, IDC_BytesPerContainer, 0, false);

#if 1

      SetDlgItemText (Dlg, IDC_ChannelConfig, L"def");

#else

      wchar_t Str [16];

      SetDlgItemText (
        Dlg, IDC_ChannelConfig, _itow (static_cast <int> (Fmt.GetCC ()), Str, 16)
      );

#endif

      FillComboBox (Dlg, IDC_MinBufferMs, MinBufferMsList);
      SendDlgItemMessage (Dlg, IDC_MinBufferMs, CB_SETCURSEL, 0, 0);

      ResetControls ();

      ShowWindow (Dlg, SW_SHOW);

      break;

    }



    case WM_COMMAND: {

      UINT const CtrlId = LOWORD (WP);
      UINT const CtrlMsg = HIWORD (WP);

      Unreferenced (CtrlMsg);

      HRESULT HR;

      if (CtrlId == IDC_SelectFile) {

        if (IsCapt ()) {

          Res = true;

          SelectInputFile ();

        } else {

          Res = true;

          SelectOutputFile ();

        }

      } else if (CtrlId == IDC_StartStop) {

        Res = true;

        if (!Started) {

          HR = Start ();

          if (FAILED (HR)) {

            Assert (!Started);
            break;

          }

          Assert (Started);

        } else { // Start/stop

          Debug (6, L"Stop command");

          Stop ();

          Assert (!Started);

        }

      } else if (CtrlId == IDCANCEL) {

        Res = true;

        Debug (6, L"Close/exit");

        //Stop ();

      }

      break;

    } // WM_COMMAND



    case WM_TIMER: {

      Debug (5, L"Timer");

      Res = true;

      if (

        Host != nullptr

#if UseLegacyWrappers

        && Host_IsStarted (Host)

#else

        && Host->IsStarted ()

#endif

      ) {

        CommonDataDesc const & CD = (

#if UseLegacyWrappers

          *Host_GetCommonData (Host)

#else

          Host->GetCommonData ()

#endif

        );

        StreamInfo const & SI = CD.Drv.SI;

        SetDlgItemText (Dlg, IDC_WorkerState, CD.Host.Ready? L"Ready" : L"Offline");
    
        SetDlgItemInt (Dlg, IDC_FramesBuffered, SI.FramesBuffered, false);
        SetDlgItemInt (Dlg, IDC_Overflows, SI.DataOverflows, false);
        SetDlgItemInt (Dlg, IDC_Underflows, SI.DataUnderflows, false);
        SetDlgItemInt (Dlg, IDC_BufferSize, SI.BufferFrames, false);

        wchar_t Text [32];

        wsprintf (Text, L"%I64u", CD.Drv.SI.FramesTransferred);
        SetDlgItemText (Dlg, IDC_FramesTransferred, Text);

        Assert (SI.State <= NumArrayElems (KsStateNames));
        SetDlgItemText (Dlg, IDC_StreamState, KsStateNames [SI.State]);

      }

      break;

    } // WM_TIMER



    // An ugly workaround for Q230587

    case WM_MOUSEACTIVATE: {

      UINT const MousePos1 = GetMessagePos ();
      POINT const MousePos2 = {MAKEPOINTS (MousePos1).x, MAKEPOINTS (MousePos1).y};
      HWND const WinUnderCursor = WindowFromPoint (MousePos2);

      Debug (
        6,
        L"Mouse activation for %#x, x=%d, y=%d (win %#x)",
        WP, MousePos2.x, MousePos2.y, WinUnderCursor
      );

      if (WinUnderCursor != 0) {

        SendMessage (Dlg, WM_NEXTDLGCTL, reinterpret_cast <WPARAM> (WinUnderCursor), true);

      }

      Verify (SetWindowPos (Dlg, HWND_TOP, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE));

      break;

    }



    default: {

      //Debug (6, L"PipeDlg - msg %#x", Msg);

      break;

    }

  }

  return Res;

}





/*
Pipe dialog function (outer)
================================================================================================
*/

INT_PTR WINAPI Worker::DialogProcOuter (HWND const Win, UINT const Msg, WPARAM const WP, LPARAM const LP) {

  //::Dbg.Printf (5, L"DlgO: win %#x, msg %#x, WP %#x, LP %#x", Win, Msg, WP, LP);

  Worker * Wrk;

  if (Msg == WM_INITDIALOG) {

    SetWindowLongPtr (Win, DWLP_USER, LP);

    Wrk = reinterpret_cast <Worker *> (LP);

    Wrk->Dlg = Win;

  } else {

    Wrk = reinterpret_cast <Worker *> (GetWindowLongPtr (Win, DWLP_USER));

  }

  INT_PTR const Res = (Wrk != nullptr)? Wrk->DialogProc (Msg, WP, LP) : false;

  return Res;

}





/*
Initialization
================================================================================================
*/

HRESULT Worker::Init () {

  HRESULT HR = E_I_PROCESSING_NOT_FINISHED;

  Assert (Dlg == 0);

  Dlg = CreateDialogParam (
    App.GetInst (), MAKEINTRESOURCE (IDD_Pipe), App.GetWindow (),
    DialogProcOuter, reinterpret_cast <LPARAM> (this)
  );

  if (Dlg == 0) {

    HR = HRESULT_FROM_WIN32 (GetLastError ());

    Error (L"Cannot create worker dialog (%s)", ErrorText (HR));

    goto Exit;

  }

  HR = S_OK;

Exit:

  Assert (HR != E_I_PROCESSING_NOT_FINISHED);

  return HR;

}





/*
Start the worker
================================================================================================
*/

HRESULT Worker::Start () {

  Assert (Thread == 0 && !ThreadExitRequest);

  HRESULT HR = E_I_PROCESSING_NOT_FINISHED;

  {

    Debug (6, L"Starting");

    EnableWindow (GetDlgItem (Dlg, IDC_StartStop), false);

    if (IsCapt ()) {

      HR = ReadDataFromFile ();

      if (FAILED (HR)) {

        goto Exit;

      }

    } else {

      Debug (6, L"Collecting data from dialog controls");

      bool Success = true;
      BOOL TmpSucc;

      UINT const SR = GetDlgItemInt (Dlg, IDC_SamplingRate, &TmpSucc, false);
      Success = Success && !!TmpSucc;

      UINT const BPS = GetDlgItemInt (Dlg, IDC_BitsPerSample, &TmpSucc, false);
      Success = Success && !!TmpSucc;

      UINT const NC = GetDlgItemInt (Dlg, IDC_NumChannels, &TmpSucc, false);
      Success = Success && !!TmpSucc;

      UINT const BPC = GetDlgItemInt (Dlg, IDC_BytesPerContainer, &TmpSucc, false);
      Success = Success && !!TmpSucc;

      if (!Success) {
        
        Error (L"Incorrect format parameter(s)");

        HR = HRESULT_FROM_WIN32 (ERROR_INVALID_PARAMETER);
        goto Exit;

      }

      wchar_t Str [16];

      GetDlgItemText (Dlg, IDC_ChannelConfig, Str, NumArrayElems (Str));

      PTSTR EndPtr = Str + 1;

      UINT const CC = (wcsicmp (Str, L"def") == 0)? UINT_MAX : wcstoul (Str, &EndPtr, 16);

      if (EndPtr == Str && *Str != L'\0') {
        
        Error (L"Incorrect %s channel config", GetDir ());

        HR = HRESULT_FROM_WIN32 (ERROR_INVALID_PARAMETER);
        goto Exit;

      }

      HR = Fmt.Init (SR, BPS, NC, BPC, CC);

      if (HR != S_OK) {

        Error (L"Incorrect %s format parameters", GetDir ());

        goto Exit;

      }

      UINT const PathLen = GetDlgItemText (Dlg, IDC_File, FilePath, NumArrayElems (FilePath));

      if (PathLen == 0) {

        Error (L"No %s file specified", IsCapt ()? L"input" : L"output");

        goto Exit;

      }

      Assert (FileData == nullptr && FileSize == 0);

      FileSize = Fmt.MsToBytes  (SecondsToRecord * 1000);

      FileData = new BYTE [FileSize];

      if (FileData == nullptr) {

        HR = HRESULT_FROM_WIN32 (ERROR_NOT_ENOUGH_MEMORY);

        Error (L"Cannot create wave file buffer");

        goto Exit;

      }

    } // capture/render

    UINT const MinBufferMsIdx = static_cast <UINT> (
      SendDlgItemMessage (Dlg, IDC_MinBufferMs, CB_GETCURSEL, 0, 0)
    );

    SetDlgItemInt (Dlg, IDC_SamplingRate, Fmt.GetSamplingRate (), false);
    SetDlgItemInt (Dlg, IDC_BitsPerSample, Fmt.GetBitsPerSample (), false);
    SetDlgItemInt (Dlg, IDC_NumChannels, Fmt.GetNumChannels (), false);
    SetDlgItemInt (Dlg, IDC_BytesPerContainer, Fmt.GetContainerSize (), false);

    {

      wchar_t Str [16];

      wsprintf (Str, L"%x", Fmt.GetChannelConfig ());

      SetDlgItemText (Dlg, IDC_ChannelConfig, Str);

    }

    UINT MinBufferMs = 0;

    if (MinBufferMsIdx != 0) {

      Assert (MinBufferMsIdx < NumArrayElems (MinBufferMsList));

      MinBufferMs = static_cast <UINT> (_wtoi (MinBufferMsList [MinBufferMsIdx]));

      Assert (MinBufferMs != 0);

    }

    FilePos = 0;

    Debug (6, L"Creating host");

    Assert (Host == nullptr);
    Assert (App.GetDriver () != nullptr);

#if UseLegacyWrappers

    HR = Host_Create (
      reinterpret_cast <void **> (&Host), App.GetDriver (), GetId (), IsCapt (), CopyDataWrapper, this
    );

#else

    HR = ThinAudioGateway::CreateHost (Host, App.GetDriver (), GetId (), IsCapt (), this);

#endif

    if (FAILED (HR)) {

      Error (L"Cannot create %s host (%s)", GetDir (), ErrorText (HR));

      goto Exit;

    }

    FormatRangeDesc const FRD = {
      Fmt.GetSamplingRate (), Fmt.GetSamplingRate (),
      Fmt.GetBitsPerSample (), Fmt.GetBitsPerSample (),
      Fmt.GetNumChannels (), Fmt.GetNumChannels (),
      Fmt.GetContainerSize (), Fmt.GetContainerSize ()
    };

    Debug (6, L"Starting host");

#if UseLegacyWrappers

    HR = Host_Start (Host, &FRD, nullptr, MinBufferMs * 1000);

#else

    HR = Host->Start (&FRD, nullptr, MinBufferMs * 1000);

#endif

    if (FAILED (HR)) {

      Error (L"Cannot start %s host (%s)", GetDir (), ErrorText (HR));

      goto Exit;

    }

    Debug (6, L"Creating thread");

    Thread = CreateThread (nullptr, 0, ThreadProcOuter, this, 0, nullptr);

    if (Thread == 0) {

      HR = HRESULT_FROM_WIN32 (GetLastError ());

      Error (L"Cannot start %s thread (%s)", GetDir (), ErrorText (HR));

      goto Exit;

    }

    Debug (7, L"Created thread %#x", Thread);

    EnableNonRealtimeControls (false);
    SetTimer (Dlg, 1, 200, nullptr);

    Started = true;

    Debug (7, L"Started");

    SetDlgItemText (Dlg, IDC_StartStop, L"Stop");
    EnableWindow (GetDlgItem (Dlg, IDC_StartStop), true);

    HR = S_OK;

  }

Exit:

  if (FAILED (HR)) {

    Stop ();

  }

  Assert (HR != E_I_PROCESSING_NOT_FINISHED);

  return HR;

}





/*
Stop the worker
================================================================================================
*/

HRESULT Worker::Stop () {

  HRESULT Res = E_I_PROCESSING_NOT_FINISHED;

  Debug (6, L"Stopping");

  EnableWindow (GetDlgItem (Dlg, IDC_StartStop), false);
  KillTimer (Dlg, 1);

#if UseLegacyWrappers

  if (Host != nullptr && Host_IsStarted (Host)) {

    Res = Host_Stop (Host);

#else

  if (Host != nullptr && Host->IsStarted ()) {

    Res = Host->Stop ();

#endif

    if (FAILED (Res)) {

      Error (L"Cannot stop %s host (%s)", GetDir (), ErrorText (Res));

    }

  }

  if (Thread != 0) {

    Assert (ThreadExitRequest == false);

    Debug (6, L"Requesting thread to exit");

    ThreadExitRequest = true;

    DWORD const WR = WaitForSingleObject (Thread, 1000);

    if (WR == WAIT_OBJECT_0) {

      Debug (6, L"Thread terminated");

    } else if (WR == WAIT_TIMEOUT) {

      Debug (9, L"Thread didn't exit, terminating it");

      TerminateThread (Thread, 9999);

    }

    Thread = 0;
    ThreadExitRequest = false;

    Res = S_OK;

  } else {

    Res = S_FALSE;

  }

  Started = false;

  if (Host != nullptr) {

#if UseLegacyWrappers

    Host_Delete (Host);

#else

    Host->Delete ();

#endif

    Host = nullptr;

    static const UINT Ctls [] = {
      IDC_BufferSize, IDC_StreamState, IDC_FramesBuffered, IDC_FramesTransferred,
      IDC_Overflows, IDC_Underflows,
    };

    for (UINT i = 0; i < NumArrayElems (Ctls); ++i) {

      SetDlgItemText (Dlg, int (Ctls [i]), L"");

    }

  }

  if (Res == S_OK && !Capture) {

    Res = WriteDataToFile ();

  }

  if (FileData != nullptr) {

    delete FileData;
    FileData = nullptr;

  }

  FileSize = 0;
  FilePos = 0;

  *FilePath = L'\0';

  ResetControls ();

  Debug (7, L"%s stopped", (Res == S_OK)? L"Fully" : L"Partially");

  return Res;

}
